#include <sys/types.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <stdint.h>

#include "common.h"
#include "macros.h"
#include "failures.h"
#include "xmalloc.h"
#include "hash.h"
#include "buffer.h"
#include "misc.h"
#include "obj.h"

struct vmlink_file {
	char		*file_name;
	unsigned char	*begin;
	size_t		file_size;

	unsigned char	*text_sect;
	unsigned int	text_sect_size;

	unsigned char	*init_sect;
	unsigned int	init_sect_size;

	unsigned int	data_sect_size;

	unsigned char	*ro_sect;
	unsigned int	ro_sect_size;

	unsigned char	*strings_sect;

	struct sym_tb_entry *symbol_table;
	unsigned int	symbol_table_size;

	struct rel_tb_entry *reloc_table;
	unsigned int	reloc_table_size;
};

extern char		*__progname;
static struct vmlink_file *vmlink_files;
static unsigned int	files_num;

static unsigned int	data_sect_size;
static struct buffer	*text_sect;
static struct buffer	*init_sect;
static struct buffer	*ro_sect;
static struct buffer	*string_sect;
static struct buffer	*symbol_sect;

static int symbol_strhash(unsigned char *str)
{
	unsigned long hash = 5381;
	int c;

	while ((c = *str++) != 0)
		hash = ((hash << 5) + hash) + c; /* hash * 33 + c */

	return hash;
}

static int symbol_strcmp(void *a, void *b)
{
	int ret;

	return_val_if_fail(a, ret_invalid);
	return_val_if_fail(b, ret_invalid);

	ret = strcmp((char *)a, (char *)b);

	return ret;
}

/* Add strings to a new string section. */
static unsigned int put_string(const char *str)
{
	unsigned int ret;
	char *c, *end;

	if (buffer_size(string_sect) == 0)
		buffer_put_string(string_sect, "");

	end = buffer_ptr(string_sect) + buffer_size(string_sect);
	c = buffer_ptr(string_sect);

	while (c < end) {
		if (!strcmp(c, str))
			return (c - (char *)buffer_ptr(string_sect));
		c += strlen(c);
		c++;
	}

	ret = buffer_size(string_sect);
	buffer_put_string(string_sect, str);

	return ret;
}

/* Helpers to fill out headers. */
static void fill_obj_header(struct obj_header *h, unsigned int shnum, unsigned int shoff, unsigned int entry)
{
	strcpy((char *)h->o_magic, ".VM");
	h->o_type = FT_EXEC;
	h->o_maj_vers = MAJOR_VERS;
	h->o_min_vers = MINOR_VERS;
	h->o_shoff = shoff;
	h->o_shnum = shnum;
	h->o_shentsize = sizeof(struct section_header);
	h->o_entry = entry;
}

static void fill_text_section_header(struct section_header *sh, unsigned int off)
{
	sh->sh_name = put_string(".text");
	sh->sh_entsize = 0;
	sh->sh_off = off;
	sh->sh_size = buffer_size(text_sect);
	sh->sh_type = SHT_TEXT;
}

static void fill_init_section_header(struct section_header *sh, unsigned int off)
{
	sh->sh_name = put_string(".init");
	sh->sh_entsize = 0;
	sh->sh_off = off;
	sh->sh_size = buffer_size(init_sect);
	sh->sh_type = SHT_INIT;
}

static void fill_data_section_header(struct section_header *sh, unsigned int off)
{
	sh->sh_name = put_string(".data");
	sh->sh_entsize = 0;
	sh->sh_off = off;
	sh->sh_size = data_sect_size;
	sh->sh_type = SHT_DATA;
}

static void fill_ro_section_header(struct section_header *sh, unsigned int off)
{
	sh->sh_name = put_string(".ro");
	sh->sh_entsize = 0;
	sh->sh_off = off;
	sh->sh_size = buffer_size(ro_sect);
	sh->sh_type = SHT_RO;
}

static void fill_sym_tb_section_header(struct section_header *sh, unsigned int off)
{
	sh->sh_name = put_string(".symbol");
	sh->sh_entsize = sizeof(struct sym_tb_entry);
	sh->sh_off = off;
	sh->sh_size = buffer_size(symbol_sect);
	sh->sh_type = SHT_SYMTAB;
}

static void fill_string_section_header(struct section_header *sh, unsigned int off)
{
	sh->sh_name = put_string(".string");
	sh->sh_entsize = 0;
	sh->sh_off = off;
	sh->sh_size = buffer_size(string_sect);
	sh->sh_type = SHT_STRTAB;
}

/* Create a new file or if it already exists -- truncate it. */
static int open_file(const char *file_name)
{
	int fd;

	fd = open(file_name, O_CREAT|O_TRUNC|O_RDWR, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	if (fd < 0)
		xerror(1, "can open file %s", file_name);

	return fd;
}

/* Write all sections, headers into a file. */
static void write_file(const char *file_name, unsigned int start_offs)
{
	struct buffer *exec_file;
	struct obj_header obj_h;
	struct section_header text_sh, init_sh, data_sh, ro_sh;
	struct section_header sym_tb_sh, string_sh;
	unsigned int offs;
	int fd;

	exec_file = buffer_new();

	offs = sizeof(struct obj_header);
	fill_text_section_header(&text_sh, offs);

	offs += buffer_size(text_sect);
	fill_init_section_header(&init_sh, offs);

	offs += buffer_size(init_sect);
	fill_data_section_header(&data_sh, offs);

	fill_ro_section_header(&ro_sh, offs);

	offs += buffer_size(ro_sect);
	fill_sym_tb_section_header(&sym_tb_sh, offs);

	offs += buffer_size(symbol_sect);
	fill_string_section_header(&string_sh, offs);
	offs += buffer_size(string_sect);

	fill_obj_header(&obj_h, 6, offs, start_offs);
	obj_put_obj_header_be(exec_file, &obj_h);

	buffer_copy(exec_file, text_sect, buffer_size(text_sect));
	buffer_copy(exec_file, init_sect, buffer_size(init_sect));
	buffer_copy(exec_file, ro_sect, buffer_size(ro_sect));
	buffer_copy(exec_file, symbol_sect, buffer_size(symbol_sect));
	buffer_copy(exec_file, string_sect, buffer_size(string_sect));

	obj_put_section_header_be(exec_file, &text_sh);
	obj_put_section_header_be(exec_file, &init_sh);
	obj_put_section_header_be(exec_file, &data_sh);
	obj_put_section_header_be(exec_file, &ro_sh);
	obj_put_section_header_be(exec_file, &sym_tb_sh);
	obj_put_section_header_be(exec_file, &string_sh);

	fd = open_file(file_name);

	buffer_write_fd(fd, exec_file);
	buffer_free(exec_file);
}

static void create_buffers()
{
	text_sect = buffer_new();
	init_sect = buffer_new();
	ro_sect = buffer_new();
	string_sect = buffer_new();
	symbol_sect = buffer_new();
}

static void release_buffers()
{
	buffer_free(text_sect);
	buffer_free(init_sect);
	buffer_free(ro_sect);
	buffer_free(string_sect);
	buffer_free(symbol_sect);
}

void print_usage()
{
	printf("Usage: %s [options] file1 file2...\n"
	" -h			print this help\n"
	" -e <symbol name>	symbol of the start execution of the program,"
				"the default is `init'\n"
	" -o <file>		place the output into <file>\n", __progname);
}

int main(int argc, char **argv)
{
	struct obj_header obj_h;
	struct section_header sh;
	struct sym_tb_entry *start, *tmp;
	struct hash_table *globals;
	struct hash_table_iter *iter;
	unsigned int i;
	unsigned char *t;

	unsigned int prev_text_size, prev_init_size;
	unsigned int prev_data_size, prev_ro_size;

	int opt;
	char *start_symbol = "init";
	char *out_file = "file.v";
	ret_t ret;

	if (argc < 2) {
		print_usage();
		exit(1);
	}

	while ((opt = getopt(argc, argv, "he:o:")) != -1) {
		switch (opt) {
		case 'h':
			print_usage();
			exit(0);
			break;
		case 'e':
			start_symbol = optarg;
			break;
		case 'o':
			out_file = optarg;
			break;
		default:
			print_usage();
			exit(1);
		}
	}

	argc -= optind;
	argv += optind;

	if (argc == 0) {
		print_usage();
		exit(0);
	}

	files_num = argc;
	vmlink_files = xmalloc(sizeof(struct vmlink_file)*files_num);

	create_buffers();

	/* Open files and map them. */
	for (i = 0; i < files_num; i++) {
		size_t len;
		int fd;

		ret = open_file_ro(argv[i], &fd);
		if (ret != ret_ok)
			xerror(1, "can't open file `%s'", argv[i]);

		ret = file_length(fd, &len);
		if (ret != ret_ok)
			xerror(1, "can't get file length `%s'", argv[i]);

		vmlink_files[i].file_size = len;

		ret = mmap_file(fd, (void **)&vmlink_files[i].begin, len);
		if (ret != ret_ok)
			xerror(1, "can't mmap file `%s'", argv[i]);

		close_file(&fd);

		vmlink_files[i].file_name = argv[i];
	}

	/* Arrange all section pointers and section sizes. */
	for (i = 0; i < files_num; i++) {
		struct buffer *obj_h_buf, *sect_head_buf;
		unsigned char *begin;
		unsigned int j;

		obj_h_buf = buffer_create_static_full(vmlink_files[i].begin,
						sizeof(struct obj_header));
		obj_get_obj_header(obj_h_buf, &obj_h);
		buffer_free(obj_h_buf);

		if (strncmp((char *)obj_h.o_magic, ".VM", 3))
			xerror(1, "bad magic in file `%s'", vmlink_files[i].file_name);

		if (obj_h.o_maj_vers != MAJOR_VERS ||
				obj_h.o_min_vers != MINOR_VERS)
			xerror(1, "incompatible file version");

		if (obj_h.o_type != FT_OBJ)
			xerror(1, "it is not an object file `%s'", vmlink_files[i].file_name);

		begin = vmlink_files[i].begin;

		for (t = vmlink_files[i].begin + obj_h.o_shoff, j = 0;
				j < obj_h.o_shnum; t += obj_h.o_shentsize, j++) {
			unsigned int size;

			sect_head_buf = buffer_create_static_full(t,
						sizeof(struct section_header));
			obj_get_section_header(sect_head_buf, &sh);
			buffer_free(sect_head_buf);

			switch (sh.sh_type) {
			case SHT_TEXT:
				vmlink_files[i].text_sect = begin + sh.sh_off;
				vmlink_files[i].text_sect_size = sh.sh_size;
				break;
			case SHT_INIT:
				vmlink_files[i].init_sect = begin + sh.sh_off;
				vmlink_files[i].init_sect_size = sh.sh_size;
				break;
			case SHT_DATA:
				vmlink_files[i].data_sect_size = sh.sh_size;
				break;
			case SHT_RO:
				vmlink_files[i].ro_sect = begin + sh.sh_off;
				vmlink_files[i].ro_sect_size = sh.sh_size;
				break;
			case SHT_STRTAB:
				vmlink_files[i].strings_sect = begin + sh.sh_off;
				break;
			case SHT_SYMTAB:
				vmlink_files[i].symbol_table = (struct sym_tb_entry *)(begin + sh.sh_off);
				size = sh.sh_size / sh.sh_entsize;
				vmlink_files[i].symbol_table_size = size;
				break;
			case SHT_RELTAB:
				vmlink_files[i].reloc_table = (struct rel_tb_entry *)(begin + sh.sh_off);
				size = sh.sh_size / sh.sh_entsize;
				vmlink_files[i].reloc_table_size = size;
				break;
			default:
				xerror(1, "unknown section type");
			}
		}
	}

	globals = hash_table_new(0, (hash_callback_t)symbol_strhash,
					(hash_compare_t)symbol_strcmp);

	/* Handle symbol table. */
	for (i = 0; i < files_num; i++) {
		unsigned int j;

		prev_data_size = data_sect_size;
		prev_text_size = buffer_size(text_sect);
		prev_init_size = buffer_size(init_sect);
		prev_ro_size = buffer_size(ro_sect);

		/* Write each section to a final section. */
		data_sect_size += vmlink_files[i].data_sect_size;
		buffer_put(text_sect, vmlink_files[i].text_sect,
						vmlink_files[i].text_sect_size);
		buffer_put(init_sect, vmlink_files[i].init_sect,
						vmlink_files[i].init_sect_size);
		buffer_put(ro_sect, vmlink_files[i].ro_sect,
						vmlink_files[i].ro_sect_size);

		for (j = 0; j < vmlink_files[i].symbol_table_size; j++) {
			struct sym_tb_entry s, *s1;
			struct buffer *sym_buf;
			char *name;

			sym_buf = buffer_create_static_full(&vmlink_files[i].symbol_table[j],
							sizeof(struct sym_tb_entry));
			obj_get_sym_tb_entry(sym_buf, &s);
			buffer_free(sym_buf);
			/* Skip undefined symbol we handle it lately. */
			if (s.st_shtype == SHT_UNDEF)
				continue;

			name = (char *)vmlink_files[i].strings_sect + s.st_name;

			s1 = xmalloc0(sizeof(*s1));
			/* Put string to new string section. */
			s1->st_name = put_string(name);
			s1->st_shtype = s.st_shtype;
			s1->st_size = s.st_size;
			s1->st_visib = s.st_visib;

			/* Calculate a right offset for the new symbol. */
			switch (s.st_shtype) {
			case SHT_DATA:
				s1->st_offset = s.st_offset + prev_data_size;
				break;
			case SHT_TEXT:
				s1->st_offset = s.st_offset + prev_text_size;
				break;
			default:
				xerror(1, "unknown section type");
			}
			obj_put_sym_tb_entry_be(symbol_sect, s1);
			/* If symbol is global we put it to the hash table
			 * we use a hash table to resolve undefined symbols
			 * lately. */
			if (s.st_visib == ST_GLOBAL) {
				ret = hash_table_insert_unique(globals, name, s1);
				if (ret != ret_ok)
					xerror(1, "symbol `%s' declared as global"
							"in several files",  name);
			} else {
				/* Free a symbol memory if it is `module'*/
				xfree(s1);
			}
		}

		/* Handle relocation records. */
		for (j = 0; j < vmlink_files[i].reloc_table_size; j++) {
			struct sym_tb_entry s;
			struct rel_tb_entry r;
			struct buffer *ptr_text, *rel_buf, *sym_buf;
			unsigned int text_size_tmp, offs, num;
			char *name;

			rel_buf = buffer_create_static_full(&vmlink_files[i].reloc_table[j],
							sizeof(struct rel_tb_entry));
			obj_get_reloc_tb_entry(rel_buf, &r);
			buffer_free(rel_buf);
			/* Choose a section and its size for patching. */
			switch (r.r_shtype) {
			case SHT_INIT:
				ptr_text = init_sect;
				text_size_tmp = prev_init_size;
				break;
			case SHT_TEXT:
				ptr_text = text_sect;
				text_size_tmp = prev_text_size;
				break;
			default:
				xerror(1, "unknowing patching section");
			}
			/* Explore a relocation record flag, offset for a `strpush'
			 * instruction was written in the section by `vmasm'. We
			 * read it from the section, calculate a new value and
			 * write to the same place. */
			switch (r.r_flag) {
			case RF_SECT:
				if (r.r_info != SHT_RO)
					xerror(1, "wait only ro section to patch");

				num = buffer_get_u32_be_at(ptr_text, r.r_offset);
				num += prev_ro_size;
				buffer_put_u32_be_at(ptr_text, num, r.r_offset);
				continue;
			case RF_SYMTB:
				/* This case handle below. */
				break;
			default:
				xerror(1, "unknown relocation flag");
			}
			/* Get the symbol information using its index in the
			 * symbol table. */
			sym_buf = buffer_create_static_full(&vmlink_files[i].symbol_table[r.r_info],
							sizeof(struct sym_tb_entry));
			obj_get_sym_tb_entry(sym_buf, &s);
			buffer_free(sym_buf);

			/* Again skip undefined symbols. */
			if (s.st_shtype == SHT_UNDEF)
				continue;

			name = (char *)vmlink_files[i].strings_sect + s.st_name;
			/* Patch the section. */
			switch (s.st_shtype) {
			case SHT_DATA:
				offs = s.st_offset + prev_data_size;
				buffer_put_u32_be_at(ptr_text, offs,
						r.r_offset+text_size_tmp);
				break;
			case SHT_TEXT:
				offs = s.st_offset + text_size_tmp;
				buffer_put_u32_be_at(ptr_text, offs,
						r.r_offset+text_size_tmp);
				break;
			default:
				xerror(1, "unknown section type");
			}
		}
	}

	prev_text_size = prev_init_size = 0;

	/* Time to resolve all undefined symbols. Look through a hash table,
	 * it helps us to resolve symbols. All symbols from the hash table are
	 * already with right offsets. */
	for (i = 0; i < files_num; i++) {
		unsigned int j;

		for (j = 0; j < vmlink_files[i].reloc_table_size; j++) {
			struct sym_tb_entry s, *s1;
			struct rel_tb_entry r;
			struct buffer *ptr_text, *rel_buf, *sym_buf;
			unsigned int text_size_tmp;
			char *name;
			ret_t ret;

			rel_buf = buffer_create_static_full(&vmlink_files[i].reloc_table[j],
							sizeof(struct rel_tb_entry));
			obj_get_reloc_tb_entry(rel_buf, &r);
			buffer_free(rel_buf);

			if (r.r_flag == RF_SECT)
				continue;
			/* Get the symbol from the original symbol table. We only
			 * need a symbol name from this table. */
			sym_buf = buffer_create_static_full(&vmlink_files[i].symbol_table[r.r_info],
							sizeof(struct sym_tb_entry));
			obj_get_sym_tb_entry(sym_buf, &s);
			buffer_free(sym_buf);

			/* Skip not undefined symbol, we already handled them. */
			if (s.st_shtype != SHT_UNDEF)
				continue;

			switch (r.r_shtype) {
			case SHT_INIT:
				ptr_text = init_sect;
				text_size_tmp = prev_init_size;
				break;
			case SHT_TEXT:
				ptr_text = text_sect;
				text_size_tmp = prev_text_size;
				break;
			default:
				xerror(1, "unknowing patching section");
			}

			name = (char *)vmlink_files[i].strings_sect + s.st_name;
			/* Resolve symbol. */
			ret = hash_table_lookup(globals, name, (void **)&s1);
			if (ret != ret_ok)
				xerror(1, "can't resolve symbol `%s'", name);

			buffer_put_u32_be_at(ptr_text, s1->st_offset,
						r.r_offset+text_size_tmp);
		}

		prev_text_size += vmlink_files[i].text_sect_size;
		prev_init_size += vmlink_files[i].init_sect_size;
	}

	/* Locate the start symbol, it must be a global symbol from the text section. */
	ret = hash_table_lookup(globals, start_symbol, (void **)&start);
	if (ret != ret_ok)
		xerror(1, "can't resolve entry point name `%s'", start_symbol);

	if (start->st_shtype != SHT_TEXT)
		xerror(1, "start symbol is not in `.text' section");

	write_file(out_file, start->st_offset);

	/* Free memory. */
	iter = hash_table_iterate_init(globals);
	while(hash_table_iterate(iter, (void **)&t, (void **)&tmp))
		xfree(tmp);

	hash_table_iterate_deinit(&iter);
	hash_table_destroy(&globals);
	release_buffers();

	return 0;
}
